#!/usr/bin/env python3

__copyright__ = "Copyright (C) 2022  Valentin Lorentz"
__license__ = "GPL-3.0-or-later"


import csv
import hashlib
import logging
import io
import multiprocessing
import sys
import time

from pathlib import Path

import requests
from tqdm import tqdm


def compute_sha1(path: Path) -> str:
    with path.open("rb") as fd:
        h = hashlib.sha1()
        while data := fd.read(40960):
            h.update(data)
    return h.hexdigest()


def download_blob(session: requests.Session, path: Path, sha1: str) -> None:
    url = f"https://archive.softwareheritage.org/api/1/content/sha1:{sha1}/raw/"
    url = f"https://webapp1.internal.softwareheritage.org/api/1/content/sha1:{sha1}/raw/"
    while True:
        resp = session.get(url, stream=True)
        if resp.status_code == 429:
            rate_limit_reset = int(resp.headers["X-RateLimit-Reset"])
            wait_seconds = max(10, rate_limit_reset - time.time())
            logging.info("Waiting timeout for %d seconds", wait_seconds)
            time.sleep(wait_seconds)
            continue
        elif resp.status_code == 200:
            break
        else:
            logging.error("Unexpected status code: %s", resp.status_code)
            logging.error(resp.text)
            sys.exit(1)

    tmp_path = path.parent / f".tmp_{sha1}"

    with tmp_path.open("wb") as fd:
        for chunk in resp.iter_content(chunk_size=40960):
            fd.write(chunk)

    if compute_sha1(tmp_path) != sha1:
        logging.error("Blob downloaded to %s does not match its checksum", tmp_path)
        sys.exit(1)

    # Atomically write to destination
    tmp_path.rename(path)
    


def download_blob_if_missing(session: requests.Session, blob_dir: Path, sha1: str) -> None:
    assert set(sha1) <= set('0123456789abcdef'), sha1
    assert len(sha1) == 40, sha1
    sharded_path = blob_dir / sha1[0:2] / sha1[2:4] / sha1
    unsharded_path = blob_dir / sha1

    for path in (sharded_path, unsharded_path):
        if path.exists():
            if compute_sha1(path) != sha1:
                logging.error("Existing blob at %s does not match its checksum", path)
                sys.exit(1)
            logging.debug("Skipping %s, already exists", sha1)
            break
    else:
        logging.debug("Downloading %s", sha1)
        download_blob(session, sharded_path, sha1)


def main(blob_dir: Path, csv_infile: Path) -> None:
    session = requests.Session()
    session.headers = {"User-Agent": "SWH License Dataset blob downloader"}

    with csv_infile.open() as fd:
        reader = csv.reader(fd)
        header = next(reader)
        if header != ['swhid', 'sha1', 'name']:
            logging.error("Unexpected header in %s: %r", csv_infile, header)
            sys.exit(1)
        for row in reader:
            try:
                (swhid, sha1, name) = row
            except ValueError:
                logging.error("Unexpected row: %r", row)
                sys.exit(1)
            download_blob_if_missing(session, blob_dir, sha1)


if __name__ == "__main__":
    logging.basicConfig(
        format="%(asctime)s : %(levelname)s : %(message)s", level=logging.INFO
    )
    if len(sys.argv) != 3:
        print("Usage: python -m licenseblobs.stats BLOB_DIR CSV_INFILE")
        print()
        print("Examples:")
        print("    python -m licenseblobs.stats blobs/ license-blobs.csv")
        print("    zstdcat license-blobs.csv.zst | python -m licenseblobs.stats blobs/ /dev/stdin")
        print()
        print("Blobs from previous versions of the dataset may be added to BLOB_DIR")
        print("before running this script, to only download new blobs")
        sys.exit(1)

    main(Path(sys.argv[1]), Path(sys.argv[2]))
