#!/usr/bin/env python3

__copyright__ = "Copyright (C) 2021  Stefano Zacchiroli"
__license__ = "GPL-3.0-or-later"


import click
import json
import logging
import multiprocessing

from pathlib import Path
from time import time

from licenseblobs import FIELD_SEP, READABLE_ENCODINGS
from licenseblobs.utils import iter_blobs, guess_mime

from scancode.api import get_licenses, get_copyrights


FIELDNAMES = [
    "sha1",
    "license",
    "score",
]

DEFAULT_MIN_SCORE = 0
DEFAULT_JOBS = 1
DEFAULT_TIMEOUT = 120
MAP_CHUNKSIZE = 1
WORKER_MAX_TASKS = 1000  # to workaround Scancode get_licenses() memory leaks

# global vars, overridden within workers (due to multiprocessing.Pool limitation that
# workers cannot be closures)
_min_score = DEFAULT_MIN_SCORE
_timeout = DEFAULT_TIMEOUT


def _detect_licenses(path):
    mime_type, encoding = guess_mime(str(path))
    license_rows = []

    res = {"sha1": path.name}

    if mime_type == "text/plain" and encoding in READABLE_ENCODINGS:
        deadline = time() + _timeout
        res["licenses"] = get_licenses(
            str(path), min_score=_min_score, deadline=deadline
        )
        license_rows = {
            (
                path.name,
                lic["spdx_license_key"],
                lic["score"], 
            )
            for lic in res["licenses"]["licenses"]
        }
        res["copyrights"] = get_copyrights(str(path))

    return (license_rows, json.dumps(res))


def _worker_setup(min_score, timeout):
    global _min_score, _timeout
    _min_score = min_score
    _timeout = timeout


def detect_licenses(
        blob_dir,
        csv_outfile,
        json_outfile,
        min_score=DEFAULT_MIN_SCORE,
        jobs=DEFAULT_JOBS,
        timeout=DEFAULT_TIMEOUT,
):
    """Detect license(s) of license blobs located under blob_dir using scancode .

    Save frequencies to csv_outfile in a 3-column (sha1, license, score) CSV format.

    """
    if csv_outfile.exists():  # ensure clean slate
        csv_outfile.unlink()

    with open(csv_outfile, "w") as csvfile, open(json_outfile, "w") as jsonfile:
        csvfile.write(FIELD_SEP.join(FIELDNAMES) + "\n")
        with multiprocessing.Pool(
            processes=jobs,
            initializer=_worker_setup,
            initargs=(min_score, timeout),
            maxtasksperchild=WORKER_MAX_TASKS,
        ) as pool:
            for (license_rows, results) in pool.imap_unordered(
                _detect_licenses,
                iter_blobs(blob_dir),
                chunksize=MAP_CHUNKSIZE,
            ):
                # each detect() call can return multiple licenses, flatten them
                for (sha1, license, score) in license_rows:
                    csvfile.write(FIELD_SEP.join([sha1, license, str(score)]) + "\n")
                assert "\n" not in results
                jsonfile.write(results + "\n")
            print("Done")


@click.command()
@click.option(
    "-s",
    "--min_score",
    help="Minimum Scancode license score",
    type=int,
    default=DEFAULT_MIN_SCORE,
    show_default=True,
)
@click.option(
    "-j",
    "--jobs",
    help="Number of parallel jobs to use",
    type=int,
    default=DEFAULT_JOBS,
    show_default=True,
)
@click.option(
    "-t",
    "--worker_timeout",
    help="Max processing time for each job, in seconds",
    type=int,
    default=DEFAULT_TIMEOUT,
    show_default=True,
)
@click.argument(
    "blob_dir",
    type=click.Path(file_okay=False, dir_okay=True, path_type=Path),
    required=True,
)
@click.argument(
    "csv_file",
    type=click.Path(file_okay=True, dir_okay=False, writable=True, path_type=Path),
    required=True,
)
@click.argument(
    "json_file",
    type=click.Path(file_okay=True, dir_okay=False, writable=True, path_type=Path),
    required=True,
)
def main(blob_dir, csv_file, json_file, min_score, jobs, worker_timeout):
    logging.basicConfig(
        format="%(asctime)s : %(levelname)s : %(message)s", level=logging.INFO
    )
    multiprocessing.set_start_method("spawn")
    detect_licenses(
        blob_dir,
        csv_file,
        json_file,
        min_score=min_score,
        jobs=jobs,
        timeout=worker_timeout,
    )


if __name__ == '__main__':
    main()
