#!/usr/bin/env python3

# SPDX-FileCopyrightText: (C) 2022 Valentin Lorentz

# SPDX-License-Identifier: GPL-3.0-or-later

import os
import re
import sys
import time

import requests

from swh.storage import get_storage  # pip3 install swh.storage


API_URL = "http://graph.internal.softwareheritage.org:5009"
STORAGE = get_storage("remote", url="http://webapp1.internal.softwareheritage.org:5002/")


if len(sys.argv) == 2 and os.path.exists(sys.argv[1]):
    with open(sys.argv[1]) as fd:
        existing_swhids = dict(
            line.strip().split("\t") for line in fd if line[-2] != "\t"
        )
elif len(sys.argv) == 1:
    existing_swhids = {}
else:
    print(
        "This command takes no argument; SWHIDs should be passed as stdin.",
        file=sys.stderr
    )
    exit(1)


#auth_token = os.environ.get("SWH_AUTH_TOKEN")
#if not auth_token:
#    print("'SWH_AUTH_TOKEN' environment variable is not defined. Set it to the value")
#    print("described in https://archive.softwareheritage.org/api/#authentication")
#    exit(1)


session = requests.Session()
session.headers["User-Agent"] = "SWH License Dataset (python/swh_whereis.py)"
#session.headers["Authorization"] = f"Bearer {auth_token}"


def retry(f, url):
    while True:
        try:
            return f(url)
        except requests.exceptions.ConnectionError:
            print(f"ConnectionError to {url} ; retrying", file=sys.stderr)
            time.sleep(60)
            continue


for line in sys.stdin:
    swhid = line.strip()
    if not swhid:
        continue
    if not swhid.startswith("swh:1:"):
        print(f"Invalid SWHID: {line}", file=sys.stderr)
        exit(1)


    # If we are running incrementally, skip the request
    origin_url = existing_swhids.get(swhid)


    if not origin_url:
        for wait_time in (0, 1, 2, 5, 10):
            time.sleep(wait_time)
            url = ( 
                f"{API_URL}/graph/leaves/{swhid}"
                f"?direction=backward&max_matching_nodes=1&return_types=ori"
            )
            resp = retry(session.get, url)
            if resp.status_code == 200:
                break
            else:
                print(file=sys.stderr)
                print(f"Status code: {resp.status_code} {url}", file=sys.stderr)
                continue
        else:
            # Failed multiple time
            resp.raise_for_status()

        content = resp.content.decode("ascii")
        if content == "\n":
            # no origin found
            origin_url = ""
        else:
            m = re.match("^swh:1:ori:([0-9a-f]{40})\s+$", content)
            assert m, f"Invalid ori SWHID: {content!r}"

            origin_sha1 = bytes.fromhex(m.group(1))

            origins = retry(STORAGE.origin_get_by_sha1, [origin_sha1])

            assert origins, "Found no origin URL for {content.strip()}"

            (origin,) = origins

            origin_url = origin["url"]

            assert " " not in origin_url, (swhid, repr(origin_url))
            assert "\n" not in origin_url, (swhid, repr(origin_url))

    print(f"{swhid}\t{origin_url}\n", end="") # end="" needed to print atomically???
